#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import struct
import sys

FMT_VERSION = "BB"

GEOB_KEY = "Serato Autotags"


def readbytes(fp: io.BytesIO | io.BufferedReader):
    for x in iter(lambda: fp.read(1), b""):
        if x == b"\00":
            break
        yield x


def parse(fp: io.BytesIO | io.BufferedReader):
    version = struct.unpack(FMT_VERSION, fp.read(2))
    assert version == (0x01, 0x01)

    for i in range(3):
        data = b"".join(readbytes(fp))
        yield float(data.decode("ascii"))


def dump(bpm: float, autogain: float, gaindb: float):
    data = struct.pack(FMT_VERSION, 0x01, 0x01)
    for value, decimals in ((bpm, 2), (autogain, 3), (gaindb, 3)):
        data += "{:.{}f}".format(value, decimals).encode("ascii")
        data += b"\x00"
    return data


if __name__ == "__main__":
    import argparse
    import configparser
    import subprocess
    import tempfile

    import mutagen._file

    from .utils.tags import get_geob, tag_geob
    from .utils.ui import get_text_editor

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("-e", "--edit", action="store_true")
    args = parser.parse_args()

    tagfile = mutagen._file.File(args.file)
    if tagfile is not None:
        fp = io.BytesIO(get_geob(tagfile, GEOB_KEY))
    else:
        fp = open(args.file, mode="rb")

    with fp:
        bpm, autogain, gaindb = parse(fp)

    if args.edit:
        editor = get_text_editor()

        with tempfile.NamedTemporaryFile() as f:
            f.write(
                (f"bpm: {bpm}\n" "autogain: {autogain}\n" "gaindb: {gaindb}\n").encode(
                    "ascii"
                )
            )
            f.flush()
            status = subprocess.call((editor, f.name))
            f.seek(0)
            output = f.read()

        if status != 0:
            error_str = (f"Command executation failed with status: {status}",)
            print(error_str, file=sys.stderr)
            raise Exception(error_str)

        cp = configparser.ConfigParser()
        try:
            cp.read_string("[Autotags]\n" + output.decode())
            bpm = cp.getfloat("Autotags", "bpm")
            autogain = cp.getfloat("Autotags", "autogain")
            gaindb = cp.getfloat("Autotags", "gaindb")
        except Exception:
            print("Invalid input, no changes made", file=sys.stderr)
            raise

        new_data = dump(bpm, autogain, gaindb)
        if tagfile:
            if tagfile is not None:
                tag_geob(tagfile, GEOB_KEY, new_data)
                tagfile.save()
        else:
            with open(args.file, mode="wb") as fp:
                fp.write(new_data)
    else:
        print(f"BPM: {bpm}")
        print(f"Auto Gain: {autogain}")
        print(f"Gain dB: {gaindb}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import enum
import io
import struct
import sys

FMT_VERSION = "BB"

GEOB_KEY = "Serato Markers_"


class EntryType(enum.IntEnum):
    INVALID = 0
    CUE = 1
    LOOP = 3


def serato32encode(data: bytes):
    """Encode 3 byte plain text into 4 byte Serato binary format."""
    a, b, c = struct.unpack("BBB", data)
    z = c & 0x7F
    y = ((c >> 7) | (b << 1)) & 0x7F
    x = ((b >> 6) | (a << 2)) & 0x7F
    w = a >> 5
    return bytes(bytearray([w, x, y, z]))


def serato32decode(data: bytes):
    """Decode 4 byte Serato binary format into 3 byte plain text."""
    w, x, y, z = struct.unpack("BBBB", data)
    c = (z & 0x7F) | ((y & 0x01) << 7)
    b = ((y & 0x7F) >> 1) | ((x & 0x03) << 6)
    a = ((x & 0x7F) >> 2) | ((w & 0x07) << 5)
    return struct.pack("BBB", a, b, c)


class Entry(object):
    FMT = ">B4sB4s6s4sBB"
    FIELDS = (
        "start_position_set",
        "start_position",
        "end_position_set",
        "end_position",
        "field5",
        "color",
        "type",
        "is_locked",
    )

    def __init__(self, *args):
        assert len(args) == len(self.FIELDS)
        for field, value in zip(self.FIELDS, args):
            setattr(self, field, value)

    def __repr__(self):
        return "{name}({data})".format(
            name=self.__class__.__name__,
            data=", ".join(
                "{}={!r}".format(name, getattr(self, name)) for name in self.FIELDS
            ),
        )

    @classmethod
    def load(cls, data: bytes):
        info_size = struct.calcsize(cls.FMT)
        info = struct.unpack(cls.FMT, data[:info_size])
        entry_data = []

        start_position_set = None
        end_position_set = None
        for field, value in zip(cls.FIELDS, info):
            if field == "start_position_set":
                assert value in (0x00, 0x7F)
                value = value != 0x7F
                start_position_set = value
            elif field == "end_position_set":
                assert value in (0x00, 0x7F)
                value = value != 0x7F
                end_position_set = value
            elif field == "start_position":
                assert start_position_set is not None
                if start_position_set:
                    value = struct.unpack(
                        ">I", serato32decode(value).rjust(4, b"\x00")
                    )[0]
                else:
                    value = None
            elif field == "end_position":
                assert end_position_set is not None
                if end_position_set:
                    value = struct.unpack(
                        ">I", serato32decode(value).rjust(4, b"\x00")
                    )[0]
                else:
                    value = None
            elif field == "color":
                value = serato32decode(value)
            elif field == "type":
                value = EntryType(value)
            entry_data.append(value)

        return cls(*entry_data)

    def dump(self):
        entry_data = []
        for field in self.FIELDS:
            value = getattr(self, field)
            if field == "start_position_set":
                value = 0x7F if not value else 0x00
            elif field == "end_position_set":
                value = 0x7F if not value else 0x00
            elif field == "color":
                value = serato32encode(value)
            elif field == "start_position":
                if value is None:
                    value = 0x7F7F7F7F
                else:
                    value = serato32encode(struct.pack(">I", value)[1:])
            elif field == "end_position":
                if value is None:
                    value = 0x7F7F7F7F
                else:
                    value = serato32encode(struct.pack(">I", value)[1:])
            elif field == "type":
                value = int(value)
            entry_data.append(value)
        return struct.pack(self.FMT, *entry_data)


class Color(Entry):
    FMT = ">4s"
    FIELDS = ("color",)


def parse(fp: io.BytesIO | io.BufferedReader):
    assert struct.unpack(FMT_VERSION, fp.read(2)) == (0x02, 0x05)

    num_entries = struct.unpack(">I", fp.read(4))[0]
    for i in range(num_entries):
        entry_data = fp.read(0x16)
        assert len(entry_data) == 0x16

        entry = Entry.load(entry_data)
        yield entry

    yield Color.load(fp.read())


def dump(new_entries: list):
    data = struct.pack(FMT_VERSION, 0x02, 0x05)
    num_entries = len(new_entries) - 1
    data += struct.pack(">I", num_entries)
    for entry_data in new_entries:
        data += entry_data.dump()
    return data


if __name__ == "__main__":
    import argparse
    import ast
    import configparser
    import math
    import subprocess
    import tempfile

    import mutagen._file

    from .utils.tags import get_geob, tag_geob
    from .utils.ui import get_hex_editor, get_text_editor, ui_ask

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("-e", "--edit", action="store_true")
    args = parser.parse_args()

    if args.edit:
        text_editor = get_text_editor()
        hex_editor = get_hex_editor()

    tagfile = mutagen._file.File(args.file)
    if tagfile is not None:
        data = get_geob(tagfile, GEOB_KEY)
    else:
        with open(args.file, mode="rb") as fp:
            data = fp.read()

    entries = list(parse(io.BytesIO(data)))
    new_entries = []
    action = None

    width = math.floor(math.log10(len(entries))) + 1
    for entry_index, entry in enumerate(entries):
        if args.edit:
            if action not in ("q", "_"):
                print("{:{}d}: {!r}".format(entry_index, width, entry))
                action = ui_ask(
                    "Edit this entry",
                    {
                        "y": "edit this entry",
                        "n": "do not edit this entry",
                        "q": (
                            "quit; do not edit this entry or any of the "
                            "remaining ones"
                        ),
                        "a": "edit this entry and all later entries in the file",
                        "b": "edit raw bytes",
                        "r": "remove this entry",
                    },
                    default="n",
                )

            if action in ("y", "a", "b"):
                while True:
                    with tempfile.NamedTemporaryFile() as f:
                        if action == "b":
                            f.write(entry.dump())
                            editor = hex_editor
                        else:
                            if action == "a":
                                entries_to_edit = (
                                    (
                                        "{:{}d}: {}".format(
                                            i,
                                            width,
                                            (
                                                repr(e.type)
                                                if e.__class__ == Entry
                                                else "Color"
                                            ),
                                        ),
                                        e,
                                    )
                                    for i, e in enumerate(
                                        entries[entry_index:], start=entry_index
                                    )
                                )
                            else:
                                entries_to_edit = (
                                    (
                                        (
                                            repr(entry.type)
                                            if entry.__class__ == Entry
                                            else "Color"
                                        ),
                                        entry,
                                    ),
                                )

                            for section, e in entries_to_edit:
                                f.write("[{}]\n".format(section).encode())
                                for field in e.FIELDS:
                                    value = getattr(e, field)
                                    if field == "type":
                                        value = int(value)
                                    f.write(
                                        "{}: {!r}\n".format(
                                            field,
                                            value,
                                        ).encode()
                                    )
                                f.write(b"\n")
                            editor = text_editor
                        f.flush()
                        status = subprocess.call((editor, f.name))
                        f.seek(0)
                        output = f.read()

                    if status != 0:
                        if (
                            ui_ask(
                                "Command failed, retry",
                                {
                                    "y": "edit again",
                                    "n": "leave unchanged",
                                },
                            )
                            == "n"
                        ):
                            break
                    else:
                        try:
                            if action != "b":
                                cp = configparser.ConfigParser()
                                cp.read_string(output.decode())
                                sections = tuple(sorted(cp.sections()))
                                if action != "a":
                                    assert len(sections) == 1

                                results = []
                                for section in sections:
                                    l, s, r = section.partition(": ")
                                    name = r if s else l
                                    cls = Color if name == "Color" else Entry
                                    e = cls(
                                        *(
                                            ast.literal_eval(
                                                cp.get(section, field),
                                            )
                                            for field in cls.FIELDS
                                        )
                                    )
                                    results.append(cls.load(e.dump()))
                            else:
                                results = [entry.load(output)]
                        except Exception as e:
                            print(str(e))
                            if (
                                ui_ask(
                                    "Content seems to be invalid, retry",
                                    {
                                        "y": "edit again",
                                        "n": "leave unchanged",
                                    },
                                )
                                == "n"
                            ):
                                break
                        else:
                            for i, e in enumerate(results, start=entry_index):
                                print("{:{}d}: {!r}".format(i, width, e))
                            subaction = ui_ask(
                                "Above content is valid, save changes",
                                {
                                    "y": "save current changes",
                                    "n": "discard changes",
                                    "e": "edit again",
                                },
                                default="y",
                            )
                            if subaction == "y":
                                new_entries.extend(results)
                                if action == "a":
                                    action = "_"
                                break
                            elif subaction == "n":
                                if action == "a":
                                    action = "q"
                                new_entries.append(entry)
                                break
            elif action in ("r", "_"):
                continue
            else:
                new_entries.append(entry)
        else:
            print("{:{}d}: {!r}".format(entry_index, width, entry))

    if args.edit:
        if new_entries == entries:
            print("No changes made.")
        else:
            new_data = dump(new_entries)

            if tagfile is not None:
                tag_geob(tagfile, GEOB_KEY, new_data)
                tagfile.save()
            else:
                with open(args.file, mode="wb") as fp:
                    fp.write(new_data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import os
import struct
import sys
from typing import Any, Callable, Generator, Iterable, Tuple, TypedDict

DATABASE_FILE = os.path.join(os.path.expanduser("~"), "Music\\_Serato_\\database V2")


class DbEntry(TypedDict):
    field: str
    field_name: str
    value: str | int | bool | list["DbEntry"]
    size_bytes: int


FIELDNAMES = {
    # Database
    "vrsn": "Version",
    "otrk": "Track",
    "ttyp": "File Type",
    "pfil": "File Path",
    "tsng": "Song Title",
    "tart": "Artist",
    "talb": "Album",
    "tgen": "Genre",
    "tlen": "Length",
    "tbit": "Bitrate",
    "tsmp": "Sample Rate",
    "tsiz": "Size",
    "tbpm": "BPM",
    "tkey": "Key",
    "tart": "Artist",
    "utme": "File Time",
    "bmis": "Missing",
    "tgrp": "Grouping",
    "tlbl": "Publisher",
    "ttyr": "Year",
    # Serato stuff
    "tadd": "Date added",
    "uadd": "Date added",
    "bbgl": "Beatgrid Locked",
    # Crates
    "osrt": "Sorting",
    "brev": "Reverse Order",
    "ovct": "Column Title",
    "tvcn": "Column Name",
    "tvcw": "Column Width",
    "ptrk": "Track Path",
}


def _get_type_id(name: str) -> str:
    # vrsn field has no type_id, but contains text
    return "t" if name == "vrsn" else name[0]


ParsedType = Tuple[str, int, Any, bytes]


def parse(fp: io.BytesIO | io.BufferedReader) -> Generator[ParsedType]:
    for header in iter(lambda: fp.read(8), b""):
        assert len(header) == 8
        name_ascii: bytes
        length: int
        name_ascii, length = struct.unpack(">4sI", header)
        name: str = name_ascii.decode("ascii")
        type_id: str = _get_type_id(name)

        data = fp.read(length)
        assert len(data) == length

        value: Any
        if type_id == "b":
            value = struct.unpack("?", data)[0]
        elif type_id in ("o", "r"):
            value = tuple(parse(io.BytesIO(data)))
        elif type_id in ("p", "t"):
            value = (data[1:] + b"\00").decode("utf-16")
        elif type_id == "s":
            value = struct.unpack(">H", data)[0]
        elif type_id == "u":
            value = struct.unpack(">I", data)[0]
        else:
            value = data

        yield name, length, value, data


class ModifyRule(TypedDict):
    field: str
    func: Callable[[str, Any], Any]


def modify(
    fp: io.BytesIO | io.BufferedWriter,
    parsed: Iterable[ParsedType],
    rules: list[ModifyRule] = [],
):
    track_filename: str = ""
    for name, length, value, data in parsed:
        name_bytes = name.encode("ascii")
        assert len(name_bytes) == 4

        type_id: str = _get_type_id(name)

        if name == "pfil":
            assert isinstance(value, str)
            track_filename = os.path.normpath(value)

        rule_has_been_done = False
        for rule in rules:
            if name == rule["field"]:
                maybe_new_value = rule["func"](track_filename, value)
                if maybe_new_value is not None:
                    assert (
                        not rule_has_been_done
                    ), f"Should only pass one rule per field (field: {name})"
                    rule_has_been_done = True
                    value = maybe_new_value

        if type_id == "b":
            data = struct.pack("?", value)
        elif type_id in ("o", "r"):
            assert isinstance(value, tuple)
            nested_buffer = io.BytesIO()
            modify(nested_buffer, value, rules)
            data = nested_buffer.getvalue()
        elif type_id in ("p", "t"):
            new_data = str(value).encode("utf-16")[2:]
            assert new_data[-1:] == b"\x00"
            data = data[:1] + new_data[:-1]
        elif type_id == "s":
            data = struct.pack(">H", value)
        elif type_id == "u":
            data = struct.pack(">I", value)

        length = len(data)

        header = struct.pack(">4sI", name_bytes, length)
        fp.write(header)
        fp.write(data)


def modify_file(file: str, rules: list[ModifyRule]):
    with open(file, "rb") as read_file:
        parsed = list(parse(read_file))

    output = io.BytesIO()
    modify(output, parsed, rules)

    with open(file, "wb") as write_file:
        output.seek(0)
        write_file.write(output.read())


def parse_to_objects(fp: io.BytesIO | io.BufferedReader | str) -> Generator[DbEntry]:
    if isinstance(fp, str):
        fp = open(fp, "rb")

    for name, length, value, data in parse(fp):
        if isinstance(value, tuple):
            try:
                new_val: list[DbEntry] = [
                    {
                        "field": n,
                        "field_name": FIELDNAMES.get(n, "Unknown"),
                        "size_bytes": l,
                        "value": v,
                    }
                    for n, l, v, d in value
                ]
            except:
                print(f"error on {value}")
                raise
            value = new_val
        else:
            value = repr(value)

        yield {
            "field": name,
            "field_name": FIELDNAMES.get(name, "Unknown"),
            "size_bytes": length,
            "value": value,
        }


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        type=argparse.FileType("rb"),
        default=open(DATABASE_FILE, "rb"),
        nargs="?",
    )
    args = parser.parse_args()

    for entry in parse_to_objects(args.file):
        if isinstance(entry["value"], list):
            print(f"{entry['field']} ({entry['field_name']}, {entry['size_bytes']} B)")
            for e in entry["value"]:
                print(
                    f"    {e['field']} ({e['field_name']}, {e['size_bytes']} B): {e['value']}"
                )
        else:
            print(
                f"{entry['field']} ({entry['field_name']}, {entry['size_bytes']} B): {entry['value']}"
            )



import sys

REPLAY_GAIN_GAIN_KEY = "replaygain_SeratoGain_gain"
REPLAY_GAIN_PEAK_KEY = "replaygain_SeratoGain_peak"


if __name__ == "__main__":
    import argparse

    import mutagen._file

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    args = parser.parse_args()

    tagfile = mutagen._file.File(args.file)
    if tagfile is not None:
        gain = tagfile.get(REPLAY_GAIN_GAIN_KEY, None)
        peak = tagfile.get(REPLAY_GAIN_PEAK_KEY, None)
        print(f"gain: {gain}")
        print(f"peak: {peak}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import collections
import io
import struct
import sys

FMT_VERSION = "BB"

GEOB_KEY = "Serato BeatGrid"

NonTerminalBeatgridMarker = collections.namedtuple(
    "NonTerminalBeatgridMarker",
    (
        "position",
        "beats_till_next_marker",
    ),
)
TerminalBeatgridMarker = collections.namedtuple(
    "TerminalBeatgridMarker",
    (
        "position",
        "bpm",
    ),
)

Footer = collections.namedtuple("Footer", ("unknown",))


def parse(fp: io.BytesIO | io.BufferedReader):
    version = struct.unpack(FMT_VERSION, fp.read(2))
    assert version == (0x01, 0x00)

    num_markers = struct.unpack(">I", fp.read(4))[0]
    for i in range(num_markers):
        position = struct.unpack(">f", fp.read(4))[0]
        data = fp.read(4)
        if i == num_markers - 1:
            bpm = struct.unpack(">f", data)[0]
            yield TerminalBeatgridMarker(position, bpm)
        else:
            beats_till_next_marker = struct.unpack(">I", data)[0]
            yield NonTerminalBeatgridMarker(position, beats_till_next_marker)

    # TODO: What's the meaning of the footer byte?
    yield Footer(struct.unpack("B", fp.read(1))[0])
    assert fp.read() == b""


if __name__ == "__main__":
    import argparse

    import mutagen._file

    from .utils.tags import get_geob

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    args = parser.parse_args()

    tagfile = mutagen._file.File(args.file)
    if tagfile is not None:
        fp = io.BytesIO(get_geob(tagfile, GEOB_KEY))
    else:
        fp = open(args.file, mode="rb")

    with fp:
        for marker in parse(fp):
            print(marker)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import base64
import os

import mutagen._file
import mutagen.aiff
import mutagen.flac
import mutagen.id3
import mutagen.mp3
import mutagen.mp4
import mutagen.oggvorbis


def get_serato_tagdata(tagfile, decode: bool = False):
    if tagfile and tagfile.tags:
        if isinstance(tagfile, (mutagen.mp3.MP3, mutagen.aiff.AIFF)):
            for tagname, tagvalue in tagfile.tags.items():
                if tagname.startswith("GEOB:Serato "):
                    yield tagname[5:], tagvalue.data
        elif isinstance(tagfile, (mutagen.flac.FLAC, mutagen.mp4.MP4)):
            for tagname, tagvalue in tagfile.tags.items():
                if not tagname.startswith("serato_") and not tagname.startswith(
                    "----:com.serato.dj:"
                ):
                    continue

                if isinstance(tagfile, mutagen.flac.FLAC):
                    encoded_data = tagvalue[0].encode("utf-8")
                else:
                    encoded_data = bytes(tagvalue[0])

                fixed_data = encoded_data

                length = len(fixed_data.splitlines()[-1])
                if length % 4 == 1:
                    fixed_data += b"A=="
                elif length % 4 == 2:
                    fixed_data += b"=="
                elif length % 4 == 3:
                    fixed_data += b"="
                data = base64.b64decode(fixed_data)

                if not data.startswith(b"application/octet-stream\0"):
                    print(f"Failed to parse tag: {tagname}")
                    continue
                fieldname_endpos = data.index(b"\0", 26)
                fieldname = data[26:fieldname_endpos].decode()
                fielddata = data[fieldname_endpos + 1 :]
                yield fieldname, fielddata if decode else encoded_data
        elif isinstance(tagfile, mutagen.oggvorbis.OggVorbis):
            for tagname, tagvalue in tagfile.tags.items():
                if not tagname.startswith("serato_"):
                    continue
                yield tagname, tagvalue[0].encode("utf-8")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("output_dir")
    parser.add_argument("input_file")
    parser.add_argument("-d", "--decode", action="store_true")
    args = parser.parse_args()

    tagfile = mutagen._file.File(args.input_file)
    for field, value in get_serato_tagdata(tagfile, decode=args.decode):
        filename = f"{field}.octet-stream"
        filepath = os.path.join(args.output_dir, filename)
        print(
            'Writing {}field "{}" to file: {}'.format(
                "decoded " if args.decode else "", field, filepath
            )
        )
        with open(filepath, mode="wb") as fp:
            fp.write(value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import struct
import sys
from typing import Generator

FMT_VERSION = "BB"

GEOB_KEY = "Serato Overview"


def parse(fp: io.BytesIO | io.BufferedReader):
    version = struct.unpack(FMT_VERSION, fp.read(2))
    assert version == (0x01, 0x05)

    for x in iter(lambda: fp.read(16), b""):
        assert len(x) == 16
        yield bytearray(x)


def draw_waveform(data: Generator[bytearray, None, None]):
    from PIL import Image, ImageColor

    img = Image.new("RGB", (240, 16), "black")
    pixels = img.load()

    for i in range(img.size[0]):
        rowdata = next(data)
        factor = len([x for x in rowdata if x < 0x80]) / len(rowdata)

        for j, value in enumerate(rowdata):
            # The algorithm to derive the colors from the data has no real
            # mathematical background and was found by experimenting with
            # different values.
            color = "hsl({hue:.2f}, {saturation:d}%, {luminance:.2f}%)".format(
                hue=(factor * 1.5 * 360) % 360,
                saturation=40,
                luminance=(value / 0xFF) * 100,
            )
            pixels[i, j] = ImageColor.getrgb(color)

    return img


if __name__ == "__main__":
    import argparse

    import mutagen._file

    from .utils.tags import get_geob

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    args = parser.parse_args()

    tagfile = mutagen._file.File(args.file)
    if tagfile is not None:
        fp = io.BytesIO(get_geob(tagfile, GEOB_KEY))
    else:
        fp = open(args.file, mode="rb")

    with fp:
        data = parse(fp)
        img = draw_waveform(data)

    img.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import ast
import base64
import configparser
import io
import struct
import sys
from typing import Any, Callable, Tuple, TypedDict

FMT_VERSION = "BB"

GEOB_KEY = "Serato Markers2"

CUE_COLORS = {
    k: bytes.fromhex(v)
    for k, v in {
        "red": "CC0000",  # Hot Cue 1
        "orange": "CC4400",
        "yelloworange": "CC8800",  # Hot Cue 2
        "yellow": "CCCC00",  # Hot Cue 4
        "limegreen1": "88CC00",
        "darkgreen": "44CC00",
        "limegreen2": "00CC00",  # Hot Cue 5
        "limegreen3": "00CC44",
        "seafoam": "00CC88",
        "cyan": "00CCCC",  # Hot Cue 7
        "lightblue": "0088CC",
        "blue1": "0044CC",
        "blue2": "0000CC",  # Hot Cue 3
        "purple1": "4400CC",
        "purple2": "8800CC",  # Hot Cue 8
        "pink": "CC00CC",  # Hot Cue 6
        "magenta": "CC0088",
        "pinkred": "CC0044",
    }.items()
}

TRACK_COLORS = {
    k: bytes.fromhex(v)
    for k, v in {
        "pink": "FF99FF",
        "darkpink": "FF99DD",
        "pinkred": "FF99BB",
        "red": "FF9999",
        "orange": "FFBB99",
        "yelloworange": "FFDD99",
        "yellow": "FFFF99",
        "limegreen1": "DDFF99",
        "limegreen2": "BBFF99",
        "limegreen3": "99FF99",
        "limegreen4": "99FFBB",
        "seafoam": "99FFDD",
        "cyan": "99FFFF",
        "lightblue": "99DDFF",
        "blue1": "99BBFF",
        "blue2": "9999FF",
        "purple": "BB99FF",
        "magenta": "DD99FF",
        "white": "FFFFFF",
        "grey": "BBBBBB",
        "black": "999999",
    }.items()
}


def readbytes(fp: io.BytesIO):
    for x in iter(lambda: fp.read(1), b""):
        if x == b"\00":
            break
        yield x


class Entry(object):
    NAME: str | None
    FIELDS: Tuple[str, ...]

    def __init__(self, *args):
        assert len(args) == len(self.FIELDS)
        for field, value in zip(self.FIELDS, args):
            setattr(self, field, value)

    def __repr__(self):
        return "{name}({data})".format(
            name=self.__class__.__name__,
            data=", ".join(
                "{}={!r}".format(name, getattr(self, name)) for name in self.FIELDS
            ),
        )

    @classmethod
    def load(cls, data: bytes):
        return cls(data)

    def dump(self) -> bytes:
        return self.data


class UnknownEntry(Entry):
    NAME = None
    FIELDS = ("data",)

    @classmethod
    def load(cls, data: bytes):
        return cls(data)

    def dump(self):
        return self.data


class BpmLockEntry(Entry):
    NAME = "BPMLOCK"
    FIELDS = ("enabled",)
    FMT = "?"

    @classmethod
    def load(cls, data: bytes):
        return cls(*struct.unpack(cls.FMT, data))

    def dump(self):
        return struct.pack(self.FMT, *(getattr(self, f) for f in self.FIELDS))


class ColorEntry(Entry):
    NAME = "COLOR"
    FMT = "c3s"
    FIELDS = (
        "field1",
        "color",
    )

    @classmethod
    def load(cls, data: bytes):
        return cls(*struct.unpack(cls.FMT, data))

    def dump(self):
        return struct.pack(self.FMT, *(getattr(self, f) for f in self.FIELDS))


class CueEntry(Entry):
    NAME = "CUE"
    FMT = ">cBIc3s2s"
    FIELDS = (
        "field1",
        "index",
        "position",
        "field4",
        "color",
        "field6",
        "name",
    )

    @classmethod
    def load(cls, data: bytes):
        info_size = struct.calcsize(cls.FMT)
        info = struct.unpack(cls.FMT, data[:info_size])
        name, nullbyte, other = data[info_size:].partition(b"\x00")
        assert nullbyte == b"\x00"
        assert other == b""
        return cls(*info, name.decode("utf-8"))

    def dump(self):
        struct_fields = self.FIELDS[:-1]
        return b"".join(
            (
                struct.pack(self.FMT, *(getattr(self, f) for f in struct_fields)),
                self.name.encode("utf-8"),
                b"\x00",
            )
        )


class LoopEntry(Entry):
    NAME = "LOOP"
    FMT = ">cBII4s4sB?"
    FIELDS = (
        "field1",
        "index",
        "startposition",
        "endposition",
        "field5",
        "field6",
        "color",
        "locked",
        "name",
    )

    @classmethod
    def load(cls, data: bytes):
        info_size = struct.calcsize(cls.FMT)
        info = struct.unpack(cls.FMT, data[:info_size])
        name, nullbyte, other = data[info_size:].partition(b"\x00")
        assert nullbyte == b"\x00"
        assert other == b""
        return cls(*info, name.decode("utf-8"))

    def dump(self):
        struct_fields = self.FIELDS[:-1]
        return b"".join(
            (
                struct.pack(self.FMT, *(getattr(self, f) for f in struct_fields)),
                self.name.encode("utf-8"),
                b"\x00",
            )
        )


class FlipEntry(Entry):
    NAME = "FLIP"
    FMT1 = "cB?"
    FMT2 = ">BI"
    FMT3 = ">BI16s"
    FIELDS = ("field1", "index", "enabled", "name", "loop", "num_actions", "actions")

    @classmethod
    def load(cls, data):
        info1_size = struct.calcsize(cls.FMT1)
        info1 = struct.unpack(cls.FMT1, data[:info1_size])
        name, nullbyte, other = data[info1_size:].partition(b"\x00")
        assert nullbyte == b"\x00"

        info2_size = struct.calcsize(cls.FMT2)
        loop, num_actions = struct.unpack(cls.FMT2, other[:info2_size])
        action_data = other[info2_size:]
        actions = []
        for i in range(num_actions):
            type_id, size = struct.unpack(cls.FMT2, action_data[:info2_size])
            action_data = action_data[info2_size:]
            if type_id == 0:
                payload = struct.unpack(">dd", action_data[:size])
                actions.append(("JUMP", *payload))
            elif type_id == 1:
                payload = struct.unpack(">ddd", action_data[:size])
                actions.append(("CENSOR", *payload))
            action_data = action_data[size:]
        assert action_data == b""

        return cls(*info1, name.decode("utf-8"), loop, num_actions, actions)

    def dump(self):
        raise NotImplementedError("FLIP entry dumps are not implemented!")


def get_entry_type(entry_name: str):
    for entry_cls in (BpmLockEntry, ColorEntry, CueEntry, LoopEntry, FlipEntry):
        if entry_cls.NAME == entry_name:
            return entry_cls
    return UnknownEntry


def parse(data: bytes):
    versionlen = struct.calcsize(FMT_VERSION)
    version = struct.unpack(FMT_VERSION, data[:versionlen])
    assert version == (0x01, 0x01)

    try:
        b64data = data[versionlen : data.index(b"\x00", versionlen)]
    except:
        b64data = data[versionlen:]
    b64data = b64data.replace(b"\n", b"")
    padding = b"A==" if len(b64data) % 4 == 1 else (b"=" * (-len(b64data) % 4))
    payload = base64.b64decode(b64data + padding)
    fp = io.BytesIO(payload)
    assert struct.unpack(FMT_VERSION, fp.read(2)) == (0x01, 0x01)
    while True:
        entry_name = b"".join(readbytes(fp)).decode("utf-8")
        if not entry_name:
            break
        entry_len = struct.unpack(">I", fp.read(4))[0]
        assert entry_len > 0

        entry_type = get_entry_type(entry_name)
        yield entry_type.load(fp.read(entry_len))


def dump(entries: list[Entry]):
    version = struct.pack(FMT_VERSION, 0x01, 0x01)

    contents = [version]
    for entry in entries:
        if entry.NAME is None:
            contents.append(entry.dump())
        else:
            data = entry.dump()
            contents.append(
                b"".join(
                    (
                        entry.NAME.encode("utf-8"),
                        b"\x00",
                        struct.pack(">I", (len(data))),
                        data,
                    )
                )
            )

    payload = b"".join(contents)
    payload_base64 = bytearray(base64.b64encode(payload).replace(b"=", b"A"))

    i = 72
    while i < len(payload_base64):
        payload_base64.insert(i, 0x0A)
        i += 73

    data = version
    data += payload_base64
    return data.ljust(470, b"\x00")


def parse_entries_file(contents: str, assert_len_1: bool):
    cp = configparser.ConfigParser()
    cp.read_string(contents)
    sections = tuple(sorted(cp.sections()))
    if assert_len_1:
        assert len(sections) == 1

    results: list[Entry] = []
    for section in sections:
        l, s, r = section.partition(": ")
        entry_type = get_entry_type(r if s else l)

        e = entry_type(
            *(
                ast.literal_eval(
                    cp.get(section, field),
                )
                for field in entry_type.FIELDS
            )
        )
        results.append(entry_type.load(e.dump()))
    return results


ValueType = bytes | str


class EntryModifyCriteria(TypedDict):
    field: str
    value_modify: Callable[[ValueType], ValueType | None]


def change_entry(
    entry: Entry,
    criteria: list[EntryModifyCriteria],
):
    output = f"[{entry.NAME}]\n"
    for field in entry.FIELDS:
        value = getattr(entry, field)

        for c in criteria:
            if field == c["field"]:
                result = c["value_modify"](value)
                if result is not None:
                    value = result

        output += f"{field}: {value!r}\n"
    output += "\n"

    entry = parse_entries_file(output, assert_len_1=True)[0]
    return entry


def is_beatgrid_locked(entries: list[Entry]):
    any(
        (isinstance(entry, BpmLockEntry) and getattr(entry, "enabled"))
        for entry in entries
    )


if __name__ == "__main__":
    import argparse
    import math
    import subprocess
    import tempfile

    import mutagen._file

    from .utils.tags import get_geob, tag_geob
    from .utils.ui import get_hex_editor, get_text_editor, ui_ask

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("-e", "--edit", action="store_true")
    args = parser.parse_args()

    if args.edit:
        text_editor = get_text_editor()
        hex_editor = get_hex_editor()

    tagfile = mutagen._file.File(args.file)
    if tagfile is not None:
        data = get_geob(tagfile, GEOB_KEY)
    else:
        with open(args.file, mode="rb") as fp:
            data = fp.read()

    entries = list(parse(data))
    new_entries: list[Entry] = []
    action = None

    width = math.floor(math.log10(len(entries))) + 1
    for entry_index, entry in enumerate(entries):
        if args.edit:
            if action not in ("q", "_"):
                print("{:{}d}: {!r}".format(entry_index, width, entry))
                action = ui_ask(
                    "Edit this entry",
                    {
                        "y": "edit this entry",
                        "n": "do not edit this entry",
                        "q": (
                            "quit; do not edit this entry or any of the "
                            "remaining ones"
                        ),
                        "a": "edit this entry and all later entries in the file",
                        "b": "edit raw bytes",
                        "r": "remove this entry",
                    },
                    default="n",
                )

            if action in ("y", "a", "b"):
                while True:
                    with tempfile.NamedTemporaryFile() as f:
                        if action == "b":
                            f.write(entry.dump())
                            editor = hex_editor
                        else:
                            if action == "a":
                                entries_to_edit = (
                                    (
                                        "{:{}d}: {}".format(i, width, e.NAME),
                                        e,
                                    )
                                    for i, e in enumerate(
                                        entries[entry_index:], start=entry_index
                                    )
                                )
                            else:
                                entries_to_edit = ((entry.NAME, entry),)

                            for section, e in entries_to_edit:
                                f.write("[{}]\n".format(section).encode())
                                for field in e.FIELDS:
                                    f.write(
                                        "{}: {!r}\n".format(
                                            field,
                                            getattr(e, field),
                                        ).encode()
                                    )
                                f.write(b"\n")
                            editor = text_editor
                        f.flush()
                        status = subprocess.call((editor, f.name))
                        f.seek(0)
                        output = f.read()

                    if status != 0:
                        if (
                            ui_ask(
                                "Command failed, retry",
                                {
                                    "y": "edit again",
                                    "n": "leave unchanged",
                                },
                            )
                            == "n"
                        ):
                            break
                    else:
                        try:
                            if action != "b":
                                results = parse_entries_file(
                                    output.decode(), assert_len_1=action != "a"
                                )
                            else:
                                results = [entry.load(output)]
                        except Exception as e:
                            print(str(e))
                            if (
                                ui_ask(
                                    "Content seems to be invalid, retry",
                                    {
                                        "y": "edit again",
                                        "n": "leave unchanged",
                                    },
                                )
                                == "n"
                            ):
                                break
                        else:
                            for i, e in enumerate(results, start=entry_index):
                                print("{:{}d}: {!r}".format(i, width, e))
                            subaction = ui_ask(
                                "Above content is valid, save changes",
                                {
                                    "y": "save current changes",
                                    "n": "discard changes",
                                    "e": "edit again",
                                },
                                default="y",
                            )
                            if subaction == "y":
                                new_entries.extend(results)
                                if action == "a":
                                    action = "_"
                                break
                            elif subaction == "n":
                                if action == "a":
                                    action = "q"
                                new_entries.append(entry)
                                break
            elif action in ("r", "_"):
                continue
            else:
                new_entries.append(entry)
        else:
            print("{:{}d}: {!r}".format(entry_index, width, entry))

    if args.edit:
        if new_entries == entries:
            print("No changes made.")
        else:
            new_data = dump(new_entries)

            if tagfile is not None:
                tag_geob(tagfile, GEOB_KEY, new_data)
                tagfile.save()
            else:
                with open(args.file, mode="wb") as fp:
                    fp.write(new_data)
